@@ -34,6 +34,18 @@ else:
 logger = logging.getLogger(__name__)
 
 
+def uses_comprehension(code):
+    """Check if given code uses comprehensions."""
+    comprehension_statements = (
+        ast.ListComp,
+        ast.SetComp,
+        ast.GeneratorExp,
+        ast.DictComp
+    )
+    nodes = ast.walk(ast.parse(code))
+    return any(isinstance(node, comprehension_statements) for node in nodes)
+
+
 class DebugWrapper(object):
     """
     Notifies the frontend when debugging starts/stops
@@ -126,26 +138,21 @@ class SpyderPdb(ipyPdb, object):  # Inherits `object` to call super() in PY2
                        " command instead")
             return
         locals = self.curframe_locals
-
-        # This is necessary to allow running comprehensions with the
-        # frame locals. It also fallbacks to the right globals if the
-        # user wants to work with them instead.
-        # See spyder-ide/spyder#13909.
-        if not 'globals()' in line:
-            ns = self.curframe.f_globals.copy()
-            ns.update(locals)
-        else:
-            ns = self.curframe.f_globals
+        globals = self.curframe.f_globals
 
         if self.pdb_use_exclamation_mark:
             # Find pdb commands executed without !
             cmd, arg, line = self.parseline(line)
             if cmd:
                 cmd_in_namespace = (
-                    cmd in ns or cmd in builtins.__dict__)
+                    cmd in globals
+                    or cmd in locals
+                    or cmd in builtins.__dict__
+                )
                 # Special case for quit and exit
                 if cmd in ("quit", "exit"):
-                    if cmd in ns and isinstance(ns[cmd], ZMQExitAutocall):
+                    if cmd in globals and isinstance(
+                            globals[cmd], ZMQExitAutocall):
                         # Use the pdb call
                         cmd_in_namespace = False
                 cmd_func = getattr(self, 'do_' + cmd, None)
@@ -184,7 +191,34 @@ class SpyderPdb(ipyPdb, object):  # Inherits `object` to call super() in PY2
                 sys.displayhook = self.displayhook
                 if execute_events:
                      get_ipython().events.trigger('pre_execute')
-                exec(code, ns, locals)
+
+                 # Mitigates a CPython bug (https://bugs.python.org/issue41918)
+                 # that prevents running comprehensions with the frame locals
+                 # in Pdb.
+                 # See https://bugs.python.org/issue21161 and
+                 # spyder-ide/spyder#13909.
+                if uses_comprehension(line):
+                    # There are three potential problems with this approach:
+                    # 1. If the code access a globals variable that is
+                    #    masked by a locals variable, it will get the locals
+                    #    one.
+                    # 2. Any edit to that variable will be lost.
+                    # 3. The globals will appear to contain all the locals
+                    #    variables.
+                    # 4. Any new locals variable will be saved to globals
+                    #    instead
+                    fake_globals = globals.copy()
+                    fake_globals.update(locals)
+                    locals_keys = locals.keys()
+                    # Don't pass locals, solves spyder-ide/spyder#16790
+                    exec(code, fake_globals)
+                    # Avoid mixing locals and globals
+                    for key in locals_keys:
+                        locals[key] = fake_globals.pop(key, None)
+                    globals.update(fake_globals)
+                else:
+                    exec(code, globals, locals)
+
                 if execute_events:
                      get_ipython().events.trigger('post_execute')
             finally:
@@ -255,8 +289,8 @@ class SpyderPdb(ipyPdb, object):  # Inherits `object` to call super() in PY2
     def stop_here(self, frame):
         """Check if pdb should stop here."""
         if (frame is not None
-                and frame.f_locals.get(
-                    "__tracebackhide__", False) == "__pdb_exit__"):
+                and "__tracebackhide__" in frame.f_locals
+                and frame.f_locals["__tracebackhide__"] == "__pdb_exit__"):
             self.onecmd('exit')
             return False
 
@@ -504,8 +538,7 @@ class SpyderPdb(ipyPdb, object):  # Inherits `object` to call super() in PY2
         Register Pdb session after reset.
         """
         super(SpyderPdb, self).reset()
-        kernel = get_ipython().kernel
-        kernel._register_pdb_session(self)
+        get_ipython().pdb_session = self
 
     def do_debug(self, arg):
         """
@@ -528,8 +561,7 @@ class SpyderPdb(ipyPdb, object):  # Inherits `object` to call super() in PY2
                 exc_info = sys.exc_info()[:2]
                 self.error(
                     traceback.format_exception_only(*exc_info)[-1].strip())
-        kernel = get_ipython().kernel
-        kernel._register_pdb_session(self)
+        get_ipython().pdb_session = self
 
     def user_return(self, frame, return_value):
         """This function is called when a return trap is set here."""
@@ -737,10 +769,10 @@ class SpyderPdb(ipyPdb, object):  # Inherits `object` to call super() in PY2
 
 def enter_debugger(filename, continue_if_has_breakpoints, code_format):
     """Enter debugger. Code format should be a format that accept filename."""
-    kernel = get_ipython().kernel
-    recursive = kernel.is_debugging()
+    shell = get_ipython()
+    recursive = shell.is_debugging()
     if recursive:
-        parent_debugger = kernel._pdb_obj
+        parent_debugger = shell.pdb_session
         sys.settrace(None)
         globals = parent_debugger.curframe.f_globals
         locals = parent_debugger.curframe_locals
@@ -770,7 +802,7 @@ def enter_debugger(filename, continue_if_has_breakpoints, code_format):
         # Reset parent debugger
         sys.settrace(parent_debugger.trace_dispatch)
         parent_debugger.lastcmd = debugger.lastcmd
-        kernel._register_pdb_session(parent_debugger)
+        shell.pdb_session = parent_debugger
     else:
         # The breakpoint might not be in the cell
         debugger.run(code)
